/*
 *  Copyright (C) 2003  Ricardo Fernndez Pascual
 *  Copyright (C) 2003  Philip Langdale
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <bonobo/bonobo-i18n.h>
#include <gtk/gtkradiomenuitem.h>
#include <gtk/gtktearoffmenuitem.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkimage.h>
#include "galeon-tab-menu.h"
#include "gul-gobject-misc.h"
#include "galeon-marshal.h"
#include "gul-string.h"
#include "gul-gui.h"

#define NOT_IMPLEMENTED g_warning ("not implemented: " G_STRLOC);
//#define DEBUG_MSG(x) g_print x
#define DEBUG_MSG(x)

/**
 * Private data
 */
struct _GaleonTabMenuPrivate {
	GtkMenuShell *ms;
	guint rebuild_timeout;
	gboolean signals_registered;

	GSList *menuitems;
	const GList *tabs;

	gint ignore_menu_callbacks;

	GHashTable *menuitem_to_tab;

	GtkWidget *parent_widget;
	GaleonWindow *window;
	GulNotebook *notebook;
};

#define REBUILD_TIMEOUT 700

/**
 * Private functions, only availble from this file
 */
static void		galeon_tab_menu_class_init		(GaleonTabMenuClass *klass);
static void		galeon_tab_menu_init			(GaleonTabMenu *e);
static void		galeon_tab_menu_finalize_impl		(GObject *o);
static void		galeon_tab_menu_build_submenu		(GaleonTabMenu *gm);
static void		galeon_tab_menu_build_item		(GaleonTabMenu *gm, 
								 GaleonTab *tab,
								 guint index);
static void		galeon_tab_menu_set_notebook		(GaleonTabMenu *tm,
								 GulNotebook *nb);
static void		galeon_tab_menu_rebuild			(GaleonTabMenu *gm);
static void 		galeon_tab_menu_rebuild_with_timeout	(GaleonTabMenu *gm);
static gboolean		galeon_tab_menu_rebuild_timeout_cb	(gpointer data);
static void		galeon_tab_menu_activate_cb		(GtkMenuItem *mi,
								 GaleonTabMenu *cm);
static void		galeon_tab_menu_notebook_tab_add_cb	(GulNotebook *nb,
								 GaleonEmbed *embed,
								 GaleonTabMenu *gm);
static void		galeon_tab_menu_notebook_tab_del_cb	(GulNotebook *nb,
								 GaleonEmbed *embed,
								 GaleonTabMenu *gm);
static void		galeon_tab_menu_notebook_destroy_cb	(GulNotebook *nb,
								 GaleonTabMenu *gm);
static void		galeon_tab_menu_embed_title_cb		(GaleonEmbed *embed,
								 GaleonTabMenu *tm);

static gpointer g_object_class;

/**
 * GtkMenu object
 */

MAKE_GET_TYPE(galeon_tab_menu, "GaleonTabMenu", GaleonTabMenu,
	      galeon_tab_menu_class_init,
	      galeon_tab_menu_init, G_TYPE_OBJECT);

static void
galeon_tab_menu_class_init(GaleonTabMenuClass *klass)
{
	G_OBJECT_CLASS(klass)->finalize = galeon_tab_menu_finalize_impl;
	g_object_class = g_type_class_peek_parent(klass);

}

static void 
galeon_tab_menu_init(GaleonTabMenu *m)
{
	GaleonTabMenuPrivate *p = g_new0(GaleonTabMenuPrivate, 1);
	m->priv = p;
	
	p->menuitem_to_tab = g_hash_table_new(NULL, NULL);
	p->ignore_menu_callbacks = 0;
	p->parent_widget = NULL;
	p->window = NULL;
	p->notebook = NULL;
}

static void
galeon_tab_menu_finalize_impl(GObject *o)
{
	GaleonTabMenu *gm = GALEON_TAB_MENU(o);
	GaleonTabMenuPrivate *p = gm->priv;
	GSList *sli;
	const GList *li;

	DEBUG_MSG (("Finalizing GaleonTabMenu\n"));

	if (p->notebook)
	{
		g_signal_handlers_disconnect_matched(p->notebook,
						     G_SIGNAL_MATCH_DATA,
						     0, 0, NULL, NULL, gm);
	}

	galeon_tab_menu_set_parent_widget(gm, NULL);
	galeon_tab_menu_set_window(gm, NULL);
	
	if (p->rebuild_timeout)
	{
		g_source_remove(p->rebuild_timeout);
	}

	for (sli = p->menuitems; sli; sli = sli->next)
	{
		g_object_unref(sli->data);
	}
	g_slist_free(p->menuitems);

	for (li = p->tabs; li; li = li->next)
	{
		GaleonEmbed *embed = galeon_tab_get_embed(li->data);
		g_signal_handlers_disconnect_matched(embed,
						     G_SIGNAL_MATCH_DATA,
						     0, 0, NULL, NULL, gm);
	}
	p->tabs = NULL;

	if (p->ms)
	{
		g_object_unref(p->ms);
	}
	g_hash_table_destroy(p->menuitem_to_tab);

	g_free(p);

	G_OBJECT_CLASS(g_object_class)->finalize (o);
}

GaleonTabMenu *
galeon_tab_menu_new(GtkMenuShell *ms)
{
	GaleonTabMenu *ret = g_object_new(GALEON_TYPE_TAB_MENU, NULL);
	GaleonTabMenuPrivate *p = ret->priv;

	p->ms = g_object_ref(ms);

	p->menuitems = NULL;
	p->tabs = NULL;
	p->window = NULL;
	p->notebook = NULL;
	galeon_tab_menu_rebuild_with_timeout(ret);

	return ret;
}

static void
galeon_tab_menu_build_submenu(GaleonTabMenu *gm)
{
	GaleonTabMenuPrivate *p = gm->priv;
	const GList *li;
	guint i=0;

	p->ignore_menu_callbacks++;

	gul_gui_setup_tearoff(p->ms);

	g_assert(p->tabs == NULL);
	g_assert(p->window);

	p->tabs = galeon_window_get_tabs(p->window);

	for (li = p->tabs; li; li = li->next, i++)
	{
		galeon_tab_menu_build_item(gm, li->data, i);
	}

	if (GTK_IS_WIDGET(p->parent_widget))
	{
		if (p->tabs)
		{
			gtk_widget_set_sensitive(p->parent_widget, TRUE);
		}
		else
		{
			gtk_widget_set_sensitive(p->parent_widget, FALSE);
		}
	}
	p->ignore_menu_callbacks--;
}

#define MENU_ITEM_MAX_LENGTH 40

static void
galeon_tab_menu_build_item(GaleonTabMenu *gm,
			   GaleonTab *tab,
			   guint index)
{
	GaleonTabMenuPrivate *p = gm->priv;
	GtkWidget *w;
	gchar *title_s = gul_string_shorten(galeon_tab_get_title(tab),
					    MENU_ITEM_MAX_LENGTH);
	gchar *label = gul_string_new_num_accel (index, title_s, TRUE);

	p->ignore_menu_callbacks++;

	w = gtk_menu_item_new_with_mnemonic(label);

	g_signal_connect(w, "activate", 
			 G_CALLBACK(galeon_tab_menu_activate_cb), gm);

	p->menuitems = g_slist_prepend(p->menuitems, g_object_ref(w));

	g_hash_table_insert(p->menuitem_to_tab, w, tab);

	gtk_widget_show(w);
	gtk_container_add(GTK_CONTAINER(p->ms), w);

	g_free(title_s);
	g_free(label);

	p->ignore_menu_callbacks--;
}

static void
galeon_tab_menu_activate_cb(GtkMenuItem *mi, GaleonTabMenu *cm)
{
	GaleonTabMenuPrivate *p = cm->priv;
	GaleonTab *tab;
	
	if (p->ignore_menu_callbacks > 0)
	{
		return;
	}

	DEBUG_MSG(("galeon_tab_menu_activated_cb\n"));

	tab = g_hash_table_lookup(p->menuitem_to_tab, mi);
	if (tab && IS_GALEON_TAB(tab))
	{
		DEBUG_MSG(("Going to '%s'\n", galeon_tab_get_title(tab)));
		
		if (p->window)
		{
			galeon_window_jump_to_tab(p->window, tab);
		}
	}
	else
	{
		DEBUG_MSG(("NULL url assoicated to menu item\n"));
	}
}

static gboolean
galeon_tab_menu_rebuild_timeout_cb(gpointer data)
{
	GaleonTabMenu *gm = data;
	GaleonTabMenuPrivate *p = gm->priv;
	DEBUG_MSG(("GaleonTabMenu rebuild timeout\n"));

	p->rebuild_timeout = 0;

	galeon_tab_menu_rebuild(data);
	return FALSE;
}

static void 
galeon_tab_menu_rebuild_with_timeout (GaleonTabMenu *gm) 
{
	GaleonTabMenuPrivate *p = gm->priv;

	DEBUG_MSG(("galeon_tab_menu_rebuild_with_timeout\n"));

	if (p->rebuild_timeout)
	{
		g_source_remove(p->rebuild_timeout);
	}
	
	p->rebuild_timeout = g_timeout_add(REBUILD_TIMEOUT, 
					   galeon_tab_menu_rebuild_timeout_cb, gm);
}

static void
galeon_tab_menu_reset (GaleonTabMenu *gm)
{
	GaleonTabMenuPrivate *p = gm->priv;
	GSList *sli;

	p->ignore_menu_callbacks++;

	/* free the list of menuitems */
	for (sli = p->menuitems; sli; sli = sli->next)
	{
		//g_assert(GTK_IS_WIDGET (sli->data) && GTK_WIDGET (sli->data)->parent == GTK_WIDGET (p->ms));
		gtk_container_remove(GTK_CONTAINER (p->ms), sli->data);
	}
	for (sli = p->menuitems; sli; sli = sli->next)
	{
		g_signal_handlers_disconnect_matched(sli->data, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, gm);
		g_object_unref(sli->data);
	}
	g_slist_free(p->menuitems);
	p->menuitems = NULL;

	p->tabs = NULL;

	if (g_hash_table_size(p->menuitem_to_tab) > 0)
	{
		g_hash_table_destroy(p->menuitem_to_tab);
		p->menuitem_to_tab = g_hash_table_new(NULL, NULL);
	}

	p->ignore_menu_callbacks--;
}

static void
galeon_tab_menu_rebuild(GaleonTabMenu *gm)
{
	GaleonTabMenuPrivate *p = gm->priv;

	DEBUG_MSG(("galeon_tab_menu_rebuild called\n"));

	p->ignore_menu_callbacks++;

	galeon_tab_menu_reset(gm);
	
	galeon_tab_menu_build_submenu(gm);

	p->ignore_menu_callbacks--;
}

void
galeon_tab_menu_set_parent_widget(GaleonTabMenu *cm, GtkWidget *w)
{
	GaleonTabMenuPrivate *p = cm->priv;
	
	if (p->parent_widget)
	{
		g_object_remove_weak_pointer(G_OBJECT (p->parent_widget),
					     (gpointer *) &p->parent_widget);
	}

	p->parent_widget = w;

	if (p->parent_widget)
	{
		g_object_add_weak_pointer(G_OBJECT (p->parent_widget), 
					  (gpointer *) &p->parent_widget);
	}
}

void
galeon_tab_menu_set_window(GaleonTabMenu *tm,
			   GaleonWindow *w)
{
	GaleonTabMenuPrivate *p = tm->priv;

	if (p->window)
	{
		g_object_remove_weak_pointer(G_OBJECT (p->window),
					     (gpointer *) &p->window);
	}

	p->window = w;

	if (p->window)
	{
		g_object_add_weak_pointer(G_OBJECT (p->window), 
					  (gpointer *) &p->window);
	}

	if (w)
	{
		galeon_tab_menu_set_notebook(tm,
					     galeon_window_get_notebook(w));
	}
}

static void
galeon_tab_menu_set_notebook(GaleonTabMenu *tm,
			     GulNotebook *nb)
{
	GaleonTabMenuPrivate *p;

	g_return_if_fail(GALEON_IS_TAB_MENU(tm));

	DEBUG_MSG(("galeon_tab_set_notebook\n"));

	p = tm->priv;

	g_return_if_fail(!p->window || IS_GALEON_WINDOW(p->window));
	g_return_if_fail(!nb || IS_GUL_NOTEBOOK(nb));

	p->notebook = nb;

	galeon_tab_menu_rebuild_with_timeout(tm);

	if (p->signals_registered) return;

	g_signal_connect(p->notebook, "tab_added",
			 G_CALLBACK(galeon_tab_menu_notebook_tab_add_cb), tm);
	g_signal_connect(p->notebook, "tab_removed",
			 G_CALLBACK(galeon_tab_menu_notebook_tab_del_cb), tm);
	g_signal_connect(p->notebook, "destroy",
			 G_CALLBACK(galeon_tab_menu_notebook_destroy_cb), tm);

	p->signals_registered = TRUE;
}

static void
galeon_tab_menu_notebook_tab_add_cb(GulNotebook *nb,
				    GaleonEmbed *embed,
				    GaleonTabMenu *tm)
{
	g_signal_connect(embed,
			 "ge_title",
			 G_CALLBACK(galeon_tab_menu_embed_title_cb), tm);
	galeon_tab_menu_rebuild_with_timeout(tm);
}

static void
galeon_tab_menu_notebook_tab_del_cb(GulNotebook *nb,
				    GaleonEmbed *embed,
				    GaleonTabMenu *tm)
{
	g_return_if_fail(embed || IS_GALEON_EMBED(embed));

	g_signal_handlers_disconnect_matched(embed,
					     G_SIGNAL_MATCH_DATA,
					     0, 0, NULL, NULL, tm);

	galeon_tab_menu_rebuild_with_timeout(tm);
}

static void
galeon_tab_menu_notebook_destroy_cb(GulNotebook *nb,
				    GaleonTabMenu *gm)
{
	GaleonTabMenuPrivate *p = gm->priv;
	const GList *li;

	g_signal_handlers_disconnect_matched(nb,
					     G_SIGNAL_MATCH_DATA,
					     0, 0, NULL, NULL, gm);

	for (li = p->tabs; li; li = li->next)
	{
		GaleonEmbed *embed = galeon_tab_get_embed(li->data);
		g_signal_handlers_disconnect_matched(embed,
						     G_SIGNAL_MATCH_DATA,
						     0, 0, NULL, NULL, gm);
        }
	p->tabs = NULL;
	p->signals_registered = FALSE;
	p->notebook = NULL;
}

static void
galeon_tab_menu_embed_title_cb(GaleonEmbed *embed,
			       GaleonTabMenu *tm)
{
	galeon_tab_menu_rebuild_with_timeout(tm);
	
}
